﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.Diagnostics;
using System.Reflection;

using System.Reflection.Emit;

using System.Linq.Expressions;

interface I1 {
	int m1(int x);
	int m2(int x);
}

interface I2 {
	int n1(int x);
	int n2(int x);
}

public class A : I1, I2 {		// Needs to be public to be accessible from the dynamically generated assembly in TestReflectionEmit()
	public int m1(int x) {
		return x + 1;
	}

	public int m2(int x) {
		return x - 1;
	}

	public int n1(int x) {
		return x + 2;
	}

	public int n2(int x) {
		return x - 2;
	}
}

abstract class X : I1, I2 {
	public abstract int m1(int x);
	public abstract int m2(int x);
	public abstract int n1(int x);
	public abstract int n2(int x);
}

class Y : X {
	public override int m1(int x) {
		return x + 1;
	}

	public override int m2(int x) {
		return x - 1;
	}

	public override int n1(int x) {
		return x + 2;
	}

	public override int n2(int x) {
		return x - 2;
	}
}

delegate int D(int v);

class Program {
	const int Iterations = 10000000;

	static int CallNonvirtual(A a) {
		return a.m1(0);
	}

	static int CallVirtual(X x) {
		return x.m1(0);
	}

	static int CallInterface(I1 i1) {
		return i1.m1(0);
	}

	static int CallDelegate(D d) {
		return d(0);
	}

	static int CallMethodInfo(A a, MethodInfo mi) {
		return (int) mi.Invoke(a, new object[] { 0 });
	}

	static int CallDynamic(dynamic d) {
		return d.m1(0);
	}

	static A TestNonvirtual() {
		A a = new A();

		int sum = 0;
		for (int i = 0; i < Iterations; i++) {
			sum += CallNonvirtual(a);
		}
		Console.WriteLine("TestNonvirtual: " + sum.ToString());

		return a;
	}

	static void TestNonvirtual2(A a) {
		int sum = 0;
		for (int i = 0; i < Iterations; i++) {
			sum += CallNonvirtual(a);
		}
		Console.WriteLine("TestNonvirtual2: " + sum.ToString());
	}

	static X TestVirtual() {
		X x = new Y();

		int sum = 0;
		for (int i = 0; i < Iterations; i++) {
			sum += CallVirtual(x);
		}
		Console.WriteLine("TestVirtual: " + sum.ToString());

		return x;
	}

	static void TestVirtual2(X x) {
		int sum = 0;
		for (int i = 0; i < Iterations; i++) {
			sum += CallVirtual(x);
		}
		Console.WriteLine("TestVirtual2: " + sum.ToString());
	}

	static I1 TestInterfaceA() {
		I1 x = new A();

		int sum = 0;
		for (int i = 0; i < Iterations; i++) {
			sum += CallInterface(x);
		}
		Console.WriteLine("TestInterfaceA (to nonvirtual): " + sum.ToString());

		return x;
	}

	static I1 TestInterfaceX() {
		I1 x = new Y();

		int sum = 0;
		for (int i = 0; i < Iterations; i++) {
			sum += CallInterface(x);
		}
		Console.WriteLine("TestInterfaceX (to virtual): " + sum.ToString());

		return x;
	}

	static void TestInterface2(I1 i1) {
		int sum = 0;
		for (int i = 0; i < Iterations; i++) {
			sum += CallInterface(i1);
		}
		Console.WriteLine("TestInterface2: " + sum.ToString());
	}

	static D TestDelegateA() {
		A a = new A();
		D d = a.m1;

		int sum = 0;
		for (int i = 0; i < Iterations; i++) {
			sum += CallDelegate(d);
		}
		Console.WriteLine("TestDelegateA (to nonvirtual): " + sum.ToString());

		return d;
	}

	static D TestDelegateX() {
		X x = new Y();
		D d = x.m1;

		int sum = 0;
		for (int i = 0; i < Iterations; i++) {
			sum += CallDelegate(d);
		}
		Console.WriteLine("TestDelegateX (to virtual): " + sum.ToString());

		return d;
	}

	static D TestDelegateIface(I1 i1) {
		D d = i1.m1;

		int sum = 0;
		for (int i = 0; i < Iterations; i++) {
			sum += CallDelegate(d);
		}
		Console.WriteLine("TestDelegateIface: " + sum.ToString());

		return d;
	}

	static void TestDelegate2(D d) {
		int sum = 0;
		for (int i = 0; i < Iterations; i++) {
			sum += CallDelegate(d);
		}
		Console.WriteLine("TestDelegate2: " + sum.ToString());
	}

	static MethodInfo TestReflection(MethodInfo mi) {
		if (mi == null) {
			mi = typeof(A).GetMethod("m1");
		}

		A a = new A();

		int sum = 0;
		for (int i = 0; i < Iterations; i++) {
			sum += CallMethodInfo(a, mi);
		}
		Console.WriteLine("TestReflection: " + sum.ToString());

		return mi;
	}

	static MethodInfo TestReflection2(MethodInfo mi) {
		if (mi == null) {
			mi = typeof(A).GetMethod("m1");
		}

		A a = new A();
		object[] args = new object[1];

		int sum = 0;
		for (int i = 0; i < Iterations; i++) {
			args[0] = 0;
			sum += (int) mi.Invoke(a, args);
		}
		Console.WriteLine("TestReflection sharing argument array: " + sum.ToString());

		return mi;
	}

	static D TestReflectionDelegate(D d) {
		if (d == null) {
			A a = new A();

			MethodInfo mi = typeof(A).GetMethod("m1");
			d = (D) Delegate.CreateDelegate(typeof(D), a, mi);
		}

		int sum = 0;
		for (int i = 0; i < Iterations; i++) {
			sum += CallDelegate(d);
		}
		Console.WriteLine("TestReflectionDelegate: " + sum.ToString());

		return d;
	}

	delegate int DCaller(A a, int x);

	static D TestCompiledExpressionTrees(D d) {
		if (d == null) {
			A a = new A();

			MethodInfo mi = typeof(A).GetMethod("m1");

			ParameterExpression param = Expression.Parameter(typeof(int), "v");
			ConstantExpression inst = Expression.Constant(a, typeof(A));
			MethodCallExpression call = Expression.Call(inst, mi, param);

			var expr = Expression.Lambda<D>(call, param);
			d = expr.Compile();
		}

		int sum = 0;
		for (int i = 0; i < Iterations; i++) {
			sum += CallDelegate(d);
		}
		Console.WriteLine("TestReflectionEmit: " + sum.ToString());

		return d;
	}

	static void TestDynamic() {
		dynamic x = new A();

		int sum = 0;
		for (int i = 0; i < Iterations; i++) {
			sum += CallDynamic(x);
		}
		Console.WriteLine("TestDynamic: " + sum.ToString());
	}

	static void TestDynamic2() {
		A x = new A();

		int sum = 0;
		for (int i = 0; i < Iterations; i++) {
			sum += CallDynamic(x);
		}
		Console.WriteLine("TestDynamic: " + sum.ToString());
	}

	static D TestReflectionEmit(D d) {
		if (d == null) {
			A a = new A();

			MethodInfo mi = typeof(A).GetMethod("m1");

			AssemblyBuilder ab = AppDomain.CurrentDomain.DefineDynamicAssembly(new AssemblyName("GeneratedAm1Caller"), AssemblyBuilderAccess.Run);
			ModuleBuilder mb = ab.DefineDynamicModule("MainModule");
			TypeBuilder tb = mb.DefineType("ACaller", TypeAttributes.Public);
			MethodBuilder methodBuilder = tb.DefineMethod("Callm1", MethodAttributes.Public | MethodAttributes.Static, typeof(int), new [] { typeof(A), typeof(int) });
			ILGenerator ilGen = methodBuilder.GetILGenerator();

			ilGen.Emit(OpCodes.Ldarg_0);
			ilGen.Emit(OpCodes.Ldarg_1);
			ilGen.Emit(OpCodes.Call, mi);
			ilGen.Emit(OpCodes.Ret);

			Type typeACaller = tb.CreateType();
			var dcaller = (DCaller) Delegate.CreateDelegate(typeof(DCaller), typeACaller.GetMethod("Callm1"));
			d = x => dcaller(a, x);
		}

		int sum = 0;
		for (int i = 0; i < Iterations; i++) {
			sum += CallDelegate(d);
		}
		Console.WriteLine("TestReflectionEmit: " + sum.ToString());

		return d;
	}

	static void Main(string[] args) {
		Console.WriteLine("Environment.Version: {0}", Environment.Version);
		Console.WriteLine("Environment.Is64BitProcess: {0}", Environment.Is64BitProcess);

		A aa = new A();
		aa.m1(5);
		aa.m2(5);
		aa.n1(5);
		aa.n2(5);

		X xx = new Y();
		xx.m1(5);
		xx.m2(5);
		xx.n1(5);
		xx.n2(5);

		GC.Collect();
		Stopwatch sw = new Stopwatch();
		{
			sw.Reset(); sw.Start();
			A a = TestNonvirtual();
			sw.Stop();
			Console.WriteLine("{0}: A warm-up time.", sw.Elapsed);

			sw.Reset(); sw.Start();
			TestNonvirtual2(a);
			sw.Stop();
			Console.WriteLine("{0}: A time.", sw.Elapsed);
		}
		Console.WriteLine();
		GC.Collect();
		{
			sw.Reset(); sw.Start();
			X x = TestVirtual();
			sw.Stop();
			Console.WriteLine("{0}: Y warm-up time.", sw.Elapsed);

			sw.Reset(); sw.Start();
			TestVirtual2(x);
			sw.Stop();
			Console.WriteLine("{0}: Y time.", sw.Elapsed);
		}
		Console.WriteLine();
		GC.Collect();
		{
			sw.Reset(); sw.Start();
			I1 i1 = TestInterfaceA();
			sw.Stop();
			Console.WriteLine("{0}: ((I1) A) warm-up time.", sw.Elapsed);

			sw.Reset(); sw.Start();
			TestInterface2(i1);
			sw.Stop();
			Console.WriteLine("{0}: ((I1) A) time.", sw.Elapsed);
		}
		Console.WriteLine();
		GC.Collect();
		{
			sw.Reset(); sw.Start();
			I1 i1 = TestInterfaceX();
			sw.Stop();
			Console.WriteLine("{0}: ((I1) Y) warm-up time.", sw.Elapsed);

			sw.Reset(); sw.Start();
			TestInterface2(i1);
			sw.Stop();
			Console.WriteLine("{0}: ((I1) Y) time.", sw.Elapsed);
		}
		Console.WriteLine();

		GC.Collect();
		{
			sw.Reset(); sw.Start();
			D d = TestDelegateA();
			sw.Stop();
			Console.WriteLine("{0}: ((D) A.m1) warm-up time.", sw.Elapsed);

			sw.Reset(); sw.Start();
			TestDelegate2(d);
			sw.Stop();
			Console.WriteLine("{0}: ((D) A.m1) time.", sw.Elapsed);
		}
		Console.WriteLine();
		GC.Collect();
		{
			sw.Reset(); sw.Start();
			D d = TestDelegateX();
			sw.Stop();
			Console.WriteLine("{0}: ((D) Y.m1) warm-up time.", sw.Elapsed);

			sw.Reset(); sw.Start();
			TestDelegate2(d);
			sw.Stop();
			Console.WriteLine("{0}: ((D) Y.m1) time.", sw.Elapsed);
		}
		Console.WriteLine();
		GC.Collect();
		{
			I1 i1 = TestInterfaceA();
			sw.Reset(); sw.Start();
			D d = TestDelegateIface(i1);
			sw.Stop();
			Console.WriteLine("{0}: ((D) ((I1) A).m1) warm-up time.", sw.Elapsed);

			sw.Reset(); sw.Start();
			TestDelegate2(d);
			sw.Stop();
			Console.WriteLine("{0}: ((D) ((I1) A).m1) time.", sw.Elapsed);
		}
		Console.WriteLine();
		GC.Collect();
		{
			I1 i1 = TestInterfaceX();
			sw.Reset(); sw.Start();
			D d = TestDelegateIface(i1);
			sw.Stop();
			Console.WriteLine("{0}: ((D) ((I1) Y).m1) warm-up time.", sw.Elapsed);

			sw.Reset(); sw.Start();
			TestDelegate2(d);
			sw.Stop();
			Console.WriteLine("{0}: ((D) ((I1) Y).m1) time.", sw.Elapsed);
		}
		Console.WriteLine();
		GC.Collect();

		Console.Write("Press ENTER to test Reflection ..."); Console.ReadLine();
		{
			I1 i1 = TestInterfaceX();
			sw.Reset(); sw.Start();
			MethodInfo mi = TestReflection(null);
			sw.Stop();
			Console.WriteLine("{0}: A via Reflection warm-up time.", sw.Elapsed);

			sw.Reset(); sw.Start();
			TestReflection(mi);
			sw.Stop();
			Console.WriteLine("{0}: A via Reflection time.", sw.Elapsed);
		}
		Console.WriteLine();
		GC.Collect();
		{
			I1 i1 = TestInterfaceX();
			sw.Reset(); sw.Start();
			MethodInfo mi = TestReflection2(null);
			sw.Stop();
			Console.WriteLine("{0}: A via Reflection (shared) warm-up time.", sw.Elapsed);

			sw.Reset(); sw.Start();
			TestReflection2(mi);
			sw.Stop();
			Console.WriteLine("{0}: A via Reflection (shared) time.", sw.Elapsed);
		}
		Console.WriteLine();
		GC.Collect();

		Console.Write("Press ENTER to test Reflection generated delegate ..."); Console.ReadLine();
		{
			I1 i1 = TestInterfaceX();
			sw.Reset(); sw.Start();
			D d = TestReflectionDelegate(null);
			sw.Stop();
			Console.WriteLine("{0}: A via delegate <- Reflection warm-up time.", sw.Elapsed);

			sw.Reset(); sw.Start();
			TestReflectionDelegate(d);
			sw.Stop();
			Console.WriteLine("{0}: A via delegate <- Reflection time.", sw.Elapsed);
		}
		Console.WriteLine();
		GC.Collect();

		Console.Write("Press ENTER to test compiled Expression Trees ..."); Console.ReadLine();
		{
			I1 i1 = TestInterfaceX();
			sw.Reset(); sw.Start();
			D d = TestCompiledExpressionTrees(null);
			sw.Stop();
			Console.WriteLine("{0}: A via Expression Trees warm-up time.", sw.Elapsed);

			sw.Reset(); sw.Start();
			TestCompiledExpressionTrees(d);
			sw.Stop();
			Console.WriteLine("{0}: A via Expression Trees time.", sw.Elapsed);
		}
		Console.WriteLine();
		GC.Collect();

		Console.Write("Press ENTER to test dynamic ..."); Console.ReadLine();
		{
			I1 i1 = TestInterfaceX();
			sw.Reset(); sw.Start();
			TestDynamic();
			sw.Stop();
			Console.WriteLine("{0}: A via dynamic (passing dynamic as dynamic) warm-up time.", sw.Elapsed);

			sw.Reset(); sw.Start();
			TestDynamic();
			sw.Stop();
			Console.WriteLine("{0}: A via dynamic (passing dynamic as dynamic) time.", sw.Elapsed);
		}
		Console.WriteLine();
		GC.Collect();
		{
			I1 i1 = TestInterfaceX();
			sw.Reset(); sw.Start();
			TestDynamic2();
			sw.Stop();
			Console.WriteLine("{0}: A via dynamic (passing A as dynamic) warm-up time.", sw.Elapsed);

			sw.Reset(); sw.Start();
			TestDynamic2();
			sw.Stop();
			Console.WriteLine("{0}: A via dynamic (passing A as dynamic) time.", sw.Elapsed);
		}
		Console.WriteLine();
		GC.Collect();

		Console.Write("Press ENTER to test Reflection.Emit ..."); Console.ReadLine();
		{
			I1 i1 = TestInterfaceX();
			sw.Reset(); sw.Start();
			D d = TestReflectionEmit(null);
			sw.Stop();
			Console.WriteLine("{0}: A via Reflection.Emit warm-up time.", sw.Elapsed);

			sw.Reset(); sw.Start();
			TestReflectionEmit(d);
			sw.Stop();
			Console.WriteLine("{0}: A via Reflection.Emit time.", sw.Elapsed);
		}
		Console.WriteLine();
		GC.Collect();
	}
}
